package checkers.igj;

import javax.annotation.processing.*;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.Element;

import com.sun.source.tree.CompilationUnitTree;
import com.sun.source.tree.ExpressionTree;
import com.sun.source.tree.Tree;

import checkers.basetype.BaseTypeChecker;
import checkers.igj.quals.*;
import checkers.metaquals.TypeQualifiers;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.AnnotatedDeclaredType;
import checkers.util.*;

/**
 * An annotation processor that checks a program use of IGJ mutability
 * type annotations, ({@code @ReadOnly}, {@code @Mutable}, and
 * {@code @Immutable}).
 *
 */
@TypeQualifiers({ AssignsFields.class, I.class, Immutable.class, Mutable.class,
    ReadOnly.class, IGJPlaceHolder.class })
@SupportedSourceVersion(SourceVersion.RELEASE_7)
public class IGJChecker extends BaseTypeChecker {

    /** Annotation factory to create annotation mirrors */
    protected AnnotationFactory annoFactory;

    /** Supported Annotations for IGJ. Used for subtyping rules **/
    protected AnnotationMirror READONLY, MUTABLE, IMMUTABLE, I, PLACE_HOLDER, ASSIGNS_FIELDS;

    protected IGJAnnotatedTypeFactory factory;

    protected static final String IMMUTABILITY_KEY = "value";

    @Override
    public synchronized void init(ProcessingEnvironment processingEnv) {
        super.init(processingEnv);
        annoFactory = new AnnotationFactory(this.env);
        READONLY = annoFactory.fromClass(ReadOnly.class);
        MUTABLE = annoFactory.fromClass(Mutable.class);
        IMMUTABLE = annoFactory.fromClass(Immutable.class);
        I = annoFactory.fromClass(I.class);
        ASSIGNS_FIELDS = annoFactory.fromClass(AssignsFields.class);
        PLACE_HOLDER = annoFactory.fromClass(IGJPlaceHolder.class);
        ((GraphQualifierHierarchy)getQualifierHierarchy()).setWildcard(PLACE_HOLDER);
    }

    @Override
    public boolean isValidUse(AnnotatedTypeMirror elemType, AnnotatedTypeMirror use) {
        if (elemType.hasAnnotation(I) || use.hasAnnotation(READONLY))
            return true;
        return super.isValidUse(elemType, use);
    }

    @Override
    protected TypeHierarchy createTypeHierarchy() {
        return new IGJTypeRelations(getQualifierHierarchy());
    }

    /**
     * Return true if the assignment variable is an assignable field or
     * variable, and returns false otherwise.
     *
     * A field is assignable if it is
     *
     * 1. a static field
     * 2. marked {@link Assignable}
     * 3. accessed through a mutable reference
     * 4. reassigned with an {@link AssignsFields} method and owned by 'this'
     *
     */
    @Override
    public boolean isAssignable(AnnotatedTypeMirror varType, Tree varTree) {
        if (varTree.getKind() == Tree.Kind.VARIABLE)
            return true;

        Element varElement = InternalUtils.symbol(varTree);
        if (varElement == null
                || !varElement.getKind().isField()
                || ElementUtils.isStatic(varElement)
                || varElement.getAnnotation(Assignable.class) != null)
            return true;

        ExpressionTree expTree = (ExpressionTree)varTree;
        AnnotatedTypeMirror receiver = factory.getReceiver(expTree);

        final boolean isAssignable =
            (receiver.hasAnnotation(MUTABLE)
             || (receiver.hasAnnotation(ASSIGNS_FIELDS) && TreeUtils.isSelfAccess(expTree)));

        return isAssignable;
    }

    @Override
    public IGJAnnotatedTypeFactory createFactory(CompilationUnitTree root) {
        factory = new IGJAnnotatedTypeFactory(this, root);
        return factory;
    }

    private final class IGJTypeRelations extends TypeHierarchy {
        public IGJTypeRelations(QualifierHierarchy qualifierHierarchy) {
            super(qualifierHierarchy);
        }

        @Override
        protected boolean compareTypeArguments(AnnotatedDeclaredType rhs, AnnotatedDeclaredType lhs) {
            if (!lhs.hasAnnotation(READONLY) && !lhs.hasAnnotation(IMMUTABLE))
                return super.compareTypeArguments(rhs, lhs);

            if (!lhs.getTypeArguments().isEmpty()
                    && !rhs.getTypeArguments().isEmpty()) {
                for (int i = 0; i < lhs.getTypeArguments().size(); ++i) {
                    if (!isSubtype(rhs.getTypeArguments().get(i), lhs.getTypeArguments().get(i)))
                        return false;
                }
            }
            return true;
        }

        @Override
        protected boolean compareTypeArg(AnnotatedTypeMirror rhs, AnnotatedTypeMirror lhs) {
            return (lhs.hasAnnotation(PLACE_HOLDER)
                    || rhs.hasAnnotation(PLACE_HOLDER)
                    || super.compareTypeArg(rhs, lhs));
        }
    }
}
