package checkers.javari;

import javax.annotation.processing.*;
import javax.lang.model.*;
import javax.lang.model.element.*;

import com.sun.source.tree.*;

import checkers.basetype.*;
import checkers.source.SuppressWarningsKey;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.AnnotatedDeclaredType;
import checkers.util.GraphQualifierHierarchy;
import checkers.javari.quals.*;
import checkers.metaquals.TypeQualifiers;
import checkers.quals.*;


/**
 * An annotation processor that checks a program's use of the Javari
 * type annotations ({@code \@ReadOnly}, {@code \@Mutable}, {@code
 * \@Assignable}, {@code \@PolyRead} and {@code \@QReadOnly}). For
 * specific instructions on the usage of each of these annotations,
 * see the Javari language specification.
 */
@SupportedSourceVersion(SourceVersion.RELEASE_7)
@SuppressWarningsKey("javari")
@TypeQualifiers( { ReadOnly.class, ThisMutable.class, Mutable.class,
    PolyRead.class, QReadOnly.class })
public class JavariChecker extends BaseTypeChecker {

    protected AnnotationMirror READONLY, THISMUTABLE, MUTABLE, POLYREAD, QREADONLY, ASSIGNABLE;

    /**
     * Initializes the checker: calls init method on super class,
     * creates a local AnnotationFactory based on the processing
     * environment, and uses it to create the protected
     * AnnotationMirrors used through this checker.
     * @param processingEnv the processing environment to use in the local AnnotationFactory
     */
    @Override
    public synchronized void init(ProcessingEnvironment processingEnv) {
        AnnotationFactory annoFactory = new AnnotationFactory(processingEnv);
        this.READONLY = annoFactory.fromClass(ReadOnly.class);
        this.THISMUTABLE = annoFactory.fromClass(ThisMutable.class);
        this.MUTABLE = annoFactory.fromClass(Mutable.class);
        this.POLYREAD = annoFactory.fromClass(PolyRead.class);
        this.QREADONLY = annoFactory.fromClass(QReadOnly.class);
        this.ASSIGNABLE = annoFactory.fromClass(Assignable.class);
        super.init(processingEnv);
    }

    @Override
    protected QualifierHierarchy createQualifierHierarchy() {
        // This should go some where else!
        GraphQualifierHierarchy annoRelations = new GraphQualifierHierarchy(READONLY);
        annoRelations.addSubtype(MUTABLE, THISMUTABLE);
        annoRelations.addSubtype(MUTABLE, null);
        annoRelations.addSubtype(null, MUTABLE);    // makes null = Mutable
        annoRelations.addSubtype(READONLY, QREADONLY);
        annoRelations.addSubtype(MUTABLE, POLYREAD);
        annoRelations.addSubtype(null, POLYREAD);
        return annoRelations;
    }

    /**
     * Implements the {@code @QReadOnly} behavior on generic types,
     * creating a new {@link TypeHierarchy} class that allows a
     * comparison of type arguments to succeed if the left hand side
     * is annotated with {@code @QReadOnly} or if the regular
     * comparison succeeds.
     */
    @Override
    protected TypeHierarchy createTypeHierarchy() {
        return new TypeHierarchy(qualHierarchy) {
            @Override
            protected boolean compareTypeArg(AnnotatedTypeMirror rhs, AnnotatedTypeMirror lhs) {
                return lhs.hasAnnotation(QREADONLY) || super.compareTypeArg(rhs, lhs);
            }
         };
    }

    /**
     * Checks if a type is primitive or if a type is subtype of another.
     */
    @Override
    public boolean isSubtype(AnnotatedTypeMirror sub, AnnotatedTypeMirror sup) {
        return sub.getKind().isPrimitive() || super.isSubtype(sub, sup);
    }


    /**
     * Always true; no type validity checking is mage by the BaseTypeVisitor.
     *
     * @see BaseTypeVisitor
     */
    @Override
    public boolean isValidUse(AnnotatedTypeMirror elemType, AnnotatedTypeMirror useType) {
        return true;
    }

    /**
     * Returns a new JavariAnnotatedTypeFactory.
     *
     * @param root the CompilationTreeUnit to use in the factory initialization.
     * @return a new JavariAnnotatedTypeFactory.
     */
    @Override
    public JavariAnnotatedTypeFactory createFactory(CompilationUnitTree root) {
        return new JavariAnnotatedTypeFactory(this, root);
    }
}
