package checkers.javari;

import java.util.*;
import javax.lang.model.element.*;
import com.sun.source.tree.*;

import checkers.source.*;
import checkers.basetype.*;
import checkers.types.*;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.TreeUtils;


/**
 * A type-checking visitor for the Javari mutability annotations
 * ({@code @ReadOnly}, {@code @Mutable} and {@code @Assignable}) that
 * extends BaseTypeVisitor.
 *
 * @see BaseTypeVisitor
 */
public class JavariVisitor extends BaseTypeVisitor<Void, Void> {

    final private AnnotationMirror READONLY, THISMUTABLE, MUTABLE,
        POLYREAD, QREADONLY, ASSIGNABLE;

    /**
     * Creates a new visitor for type-checking the Javari mutability
     * annotations.
     *
     * @param checker the {@link JavariChecker} to use
     * @param root the root of the input program's AST to check
     */
    public JavariVisitor(JavariChecker checker, CompilationUnitTree root) {
        super(checker, root);
        this.checker = checker;
        READONLY = checker.READONLY;
        THISMUTABLE = checker.THISMUTABLE;
        MUTABLE = checker.MUTABLE;
        POLYREAD = checker.POLYREAD;
        QREADONLY = checker.QREADONLY;
        ASSIGNABLE = checker.ASSIGNABLE;
        assert atypeFactory instanceof JavariAnnotatedTypeFactory : atypeFactory;
    }

    /**
     * Ensures the class type is valid.
     */
    @Override
    public Void visitClass(ClassTree node, Void p) {
        if (atypeFactory.fromClass(node).hasAnnotation(POLYREAD)
            && !atypeFactory.getSelfType(node).hasAnnotation(POLYREAD))
            checker.report(Result.failure("polyread.type"), node);

        return super.visitClass(node, p);
    }

    /**
     * Overrides its super method, calling the assignment check unless
     * the error key is generic.argument.invalid and the upper bound
     * is java.lang.Object .
     */
    @Override
    protected void commonAssignmentCheck(AnnotatedTypeMirror varType,
            AnnotatedTypeMirror valueType, Tree valueTree, String errorKey, Void p) {
        if (errorKey.equals("generic.argument.invalid")
            && varType.getUnderlyingType().toString().equals("java.lang.Object"))
            return;

        super.commonAssignmentCheck(varType, valueType, valueTree, errorKey, p);
    }

    /**
     * Checks whether the local field of a readonly reference is being
     * assigned, before calling commonAssignmentCheck to proceed with
     * subtype checking.
     */
    @Override
    protected void commonAssignmentCheck(Tree varTree, ExpressionTree valueTree, String errorKey, Void p) {

        AnnotatedTypeMirror varType = atypeFactory.getAnnotatedType(varTree),
            valueType = atypeFactory.getAnnotatedType(valueTree);

        AnnotatedDeclaredType mReceiver = atypeFactory.getSelfType(varTree);

        boolean variableLocalField;

        // we want to determine if the variable is a field of the
        // local class, so if the current class has a field f, a.f is
        // not a local field, while this.f and f are local fields.
        if (varTree.getKind() == Tree.Kind.MEMBER_SELECT) {
            AnnotatedTypeMirror oType = atypeFactory.getAnnotatedType(((MemberSelectTree)varTree).getExpression());
            variableLocalField = varType != null && varType.getElement() != null
                && varType.getElement().getKind().isField()
                && atypeFactory.getSelfType(varTree).getElement().equals(oType.getElement());
        } else {
            if (varType == null || varType.getElement() == null)
                variableLocalField = false;
            else
                variableLocalField = varType.getElement().getKind().isField();
        }

        boolean inConstructor = inConstructor();

        if (mReceiver != null)
            if ((mReceiver.hasAnnotation(READONLY)
                 || mReceiver.hasAnnotation(POLYREAD))
                && !inConstructor)
                if (!varType.hasAnnotation(ASSIGNABLE) && variableLocalField)
                    checker.report(Result.failure("ro.field"), varTree);

        if (varTree.getKind() == Tree.Kind.MEMBER_SELECT) {
            ExpressionTree et = ((MemberSelectTree)varTree).getExpression();
            AnnotatedTypeMirror oType = atypeFactory.getAnnotatedType(et);
            Element exprElt = oType.getElement();
            atypeFactory.postAsMemberOf(varType, oType, varType.getElement());

            if (!(variableLocalField && inConstructor)
                && (exprElt == null || exprElt.getEnclosingElement().getKind() != ElementKind.METHOD)) {
                if (!varType.hasAnnotation(ASSIGNABLE)
                    && (oType.hasAnnotation(READONLY)
                        || oType.hasAnnotation(QREADONLY)
                        || oType.hasAnnotation(POLYREAD)))
                    checker.report(Result.failure("ro.field"),varTree);

            }
        }

        commonAssignmentCheck(varType, valueType, valueTree, errorKey, p);

    }

    /**
     * Checks for prefix and postfix increments and decrements on
     * unassignable fields of readonly classes.
     *
     * @see com.sun.source.util.TreeScanner#visitUnary(com.sun.source.tree.UnaryTree,java.lang.Object)
     */
    @Override
    public Void visitUnary(UnaryTree node, Void p){
        Tree.Kind kind = node.getKind();
        AnnotatedTypeMirror rType = atypeFactory.getReceiver(node.getExpression());
        if (rType != null && !rType.hasAnnotation(MUTABLE) && !inConstructor()) {
            if (kind == Tree.Kind.POSTFIX_DECREMENT
                || kind == Tree.Kind.POSTFIX_INCREMENT
                || kind == Tree.Kind.PREFIX_DECREMENT
                || kind == Tree.Kind.PREFIX_INCREMENT) {

                AnnotatedTypeMirror variable
                    = atypeFactory.getAnnotatedType(node.getExpression());

                if (!variable.hasAnnotation(ASSIGNABLE)
                        && variable.getElement().getKind().isField())
                    checker.report(Result.failure("ro.field"),
                                   node.getExpression());
            }
        }
        return super.visitUnary(node, p);
    }


    /**
     * Tests whether the tree expressed by the passed type tree
     * contains a qualified primitive type on its qualified type, and
     * if so emits an error.
     *
     * @param tree  the AST type supplied by the user
     */
    @Override
    public void validateTypeOf(Tree tree) {
        Tree tTree;
        if (tree.getKind() == Tree.Kind.VARIABLE) {
            tTree = ((VariableTree)tree).getType();
            switch (tTree.getKind()) {
            case PRIMITIVE_TYPE: {
                AnnotatedTypeMirror vType = atypeFactory.getAnnotatedType(tree);
                if (vType.hasAnnotation(QREADONLY)
                    || vType.hasAnnotation(READONLY)
                    || vType.hasAnnotation(POLYREAD))
                    checker.report(Result.failure("primitive.ro"), tree);
                break;
            }
            case PARAMETERIZED_TYPE: {
                ParameterizedTypeTree pt = (ParameterizedTypeTree)tTree;
                for (Tree t : pt.getTypeArguments()) validateTypeOf(t);
                break;
            }
            }
        } else
            tTree = tree;

        switch (tTree.getKind()) {
        case ANNOTATED_TYPE: {
            AnnotatedTypeMirror vType = atypeFactory.getAnnotatedType(tree);
            Tree nextTree = ((AnnotatedTypeTree)tTree).getUnderlyingType();
            if (nextTree.getKind() == Tree.Kind.PRIMITIVE_TYPE) {
                if (vType.hasAnnotation(QREADONLY)
                    || vType.hasAnnotation(READONLY)
                    || vType.hasAnnotation(POLYREAD)) {
                    checker.report(Result.failure("primitive.ro"), tTree);
                }
            } else {
                validateTypeOf(nextTree);
            }
            break;
        }
        case ARRAY_TYPE: {
            validateTypeOf(((ArrayTypeTree)tTree).getType());
            break;
        }
        }
        super.validateTypeOf(tree);
    }


    /**
     * Emits a warning if a cast is made increasing mutability.
     *
     * @see com.sun.source.util.TreeScanner#visitTypeCast(com.sun.source.tree.TypeCastTree,java.lang.Object)
     */
    @Override
    public Void visitTypeCast(TypeCastTree node, Void p) {
        AnnotatedTypeMirror exprType = atypeFactory.getAnnotatedType(node.getExpression()),
            castType = atypeFactory.getAnnotatedType(node);

        if (!checker.isSubtype(exprType, castType)) {
            checker.report(Result.warning("mutable.cast"), node.getType());
        }

        return super.visitTypeCast(node, p);
    }

    /**
     * Overriding invocability check to let {code @ReadOnly}
     * references invoke {code @PolyRead} methods.
     */
    @Override
    protected boolean checkMethodInvocability(AnnotatedExecutableType method, MethodInvocationTree node) {
        AnnotatedTypeMirror methodReceiver = method.getReceiverType().getErased();
        AnnotatedTypeMirror treeReceiver;

        if (methodReceiver.hasAnnotation(POLYREAD)) {
            methodReceiver.addAnnotation(READONLY);
            methodReceiver.removeAnnotation(POLYREAD);
        }

        if (node.getMethodSelect().getKind() == Tree.Kind.MEMBER_SELECT) {
            MemberSelectTree mst = (MemberSelectTree) (node.getMethodSelect());
            treeReceiver = atypeFactory.getAnnotatedType(mst.getExpression());
        } else {
            treeReceiver = atypeFactory.getSelfType(node);
        }

        if (!checker.isSubtype(treeReceiver, methodReceiver)) {
            checker.report(Result.failure("ro.reference",
                TreeUtils.elementFromUse(node),
                treeReceiver.toString(), methodReceiver.toString()), node);
            return false;
        }
        return true;
    }


    /**
     * Overriding assignment check to annotate every {code @PolyRead} parameter with {code @ReadOnly}.
     */
    @Override
    protected void checkArguments(List<? extends AnnotatedTypeMirror> requiredArgs,
            List<? extends ExpressionTree> passedArgs, Void p) {

        for (int i = 0; i < requiredArgs.size(); ++i) {
            AnnotatedTypeMirror pType = requiredArgs.get(i);
            if (pType.hasAnnotation(POLYREAD) && !pType.hasAnnotation(MUTABLE)) {
                pType.addAnnotation(READONLY);
                pType.removeAnnotation(POLYREAD);
            } else if (!pType.hasAnnotation(READONLY)) {
                pType.addAnnotation(MUTABLE);
            }

            commonAssignmentCheck(pType,
                    passedArgs.get(i),
                    "argument.invalid", p);
        }
    }


    private boolean inConstructor() {
        MethodTree mt = visitorState.getMethodTree();
        if (mt == null) return true;
        ExecutableElement mElt = TreeUtils.elementFromDeclaration(mt);
        return mElt.getKind() == ElementKind.CONSTRUCTOR;
    }

}
