package checkers.nonnull;

import javax.lang.model.element.*;

import checkers.basetype.*;
import checkers.nonnull.quals.NonNull;
import checkers.source.Result;
import checkers.types.AnnotatedTypeMirror;
import checkers.types.AnnotatedTypeMirror.*;
import checkers.util.TreeUtils;

import com.sun.source.tree.*;

/**
 * A type-checking visitor for the {@link checkers.nonnull.quals.NonNull} type
 * qualifier. This visitor reports errors or warnings for violations for
 * the following cases:
 *
 * 1. if the receiver of a member dereference is not NonNull (error
 *    "dereference.of.nullable")
 * 2. if the receiver of an implicit ".iterator()" dereference in an enhanced
 *    for loop is not NonNull (error "dereference.of.nullable")
 *
 * Self-accesses (via {@code this} or {@code super}) can always be dereferenced.
 */
public class NonNullVisitor extends BaseTypeVisitor<Void, Void> {

    /** The {@link NonNull} annotation */
    private final AnnotationMirror NONNULL;

    /**
     * Creates a new visitor for type-checking {@link checkers.nonnull.quals.NonNull}.
     *
     * @param checker the checker to use
     * @param root the root of the input program's AST to check
     */
    public NonNullVisitor(NonNullChecker checker, CompilationUnitTree root) {
        super(checker, root);
        NONNULL = this.annoFactory.fromClass(NonNull.class);
    }

    @Override
    public Void visitMemberSelect(MemberSelectTree node, Void p) {
        AnnotatedTypeMirror type = atypeFactory.getAnnotatedType(node.getExpression());
        if (!TreeUtils.isSelfAccess(node))
            checkForNullability(type, node.getExpression());

        return super.visitMemberSelect(node, p);
    }

    @Override
    public Void visitEnhancedForLoop(EnhancedForLoopTree node, Void p) {
        AnnotatedTypeMirror iterableType =
            atypeFactory.getAnnotatedType(node.getExpression());
        checkForNullability(iterableType, node.getExpression());
        return super.visitEnhancedForLoop(node, p);
    }

    @Override
    protected boolean checkMethodInvocability(AnnotatedExecutableType method,
            MethodInvocationTree node) {
        // Claim that methods with a @NonNull receiver are invokable so that
        // visitMemberSelect issues dereference errors instead.
        if (!TreeUtils.isSelfAccess(node)
                && method.getReceiverType().hasAnnotation(NONNULL))
            return true;
        return super.checkMethodInvocability(method, node);
    }

    /**
     * Issues a 'dereference.of.nullable' if the type is not of a
     * {@link NonNull} type.
     *
     * @param type  type to be checked nullability
     * @param tree  the tree where the error is to reported
     */
    private void checkForNullability(AnnotatedTypeMirror type, Tree tree) {
        if (!type.hasAnnotation(NONNULL))
            checker.report(Result.failure("dereference.of.nullable", tree), tree);
    }
}
