package checkers.types;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Set;

import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.Name;

import checkers.util.AnnotationUtils;

/**
 * Represents the relationship hierarchy of some given type qualifiers.
 *
 * All method parameter annotations need to be type qualifiers recognized
 * within this hierarchy.
 */
public abstract class QualifierHierarchy {

    /**
     * Tests whether anno1 is a super qualifier anno2, according to the type
     * qualifiers hierarchy.
     * @param anno2    a type qualifier
     * @param anno1    a type qualifier
     *
     * @return true iff anno1 is a sub qualifier of anno2
     */
    public abstract boolean isSubtype(AnnotationMirror anno1, AnnotationMirror anno2);

    /**
     * Check if the two annotations are in conflict.
     *
     * Two qualifiers are in conflict if both cannot annotate a type together.
     *
     * @param anno1     a type qualifier
     * @param anno2     a type qualifier
     * @return true iff anno1 and anno2 in conflict
     */
    public abstract boolean isInConflict(AnnotationMirror anno1, AnnotationMirror anno2);

    /**
     * @return  the root (ultimate super) type qualifier in the hierarchy
     */
    public abstract AnnotationMirror getRootAnnotation();

    /**
     * Finds the type qualifiers recognized within this hierarchy in the passed
     * collections.
     *
     * @param annos     a collection of annotations
     * @return  the qualifiers in annos within this hierarchy
     */
    public List<AnnotationMirror> getAnnotations(Collection<AnnotationMirror> annos) {
        List<AnnotationMirror> results = new ArrayList<AnnotationMirror>();
        Set<Name> typeQualifiers = getTypeQualifiers();
        for (AnnotationMirror anno : annos) {
            if (typeQualifiers.contains(AnnotationUtils.annotationName(anno)))
                results.add(anno);
        }
        return results;
    }

    /**
     * Finds the first type qualifiers in this hierarchy in the given list of
     * qualifiers.
     *
     * @param annos     a collection of annotations
     * @return  the qualifiers in annos in this hierarchy
     */
    public AnnotationMirror getAnnotation(Collection<AnnotationMirror> annos) {
        Set<Name> typeQualifiers = getTypeQualifiers();
        for (AnnotationMirror anno : annos) {
            if (typeQualifiers.contains(AnnotationUtils.annotationName(anno)))
                return anno;
        }
        return null;
    }

    /**
     * Returns the names of the type qualifiers recognized within this type
     * qualifier hierarchy
     *
     * @return the fully qualified name represented in this hierarchy
     */
    public abstract Set<Name> getTypeQualifiers();

    /**
     * Tests whether there is any annotation in lhs that is a super qualifier
     * of an annotation in rhs.
     *
     * @return true iff an annotation in lhs is a super of one in rhs
     **/
    public boolean isSubtype(Collection<AnnotationMirror> rhs, Collection<AnnotationMirror> lhs) {
        for (AnnotationMirror lhsAnno : lhs)
            for (AnnotationMirror rhsAnno : rhs)
                if (isSubtype(rhsAnno, lhsAnno))
                    return true;
        return false;
    }
}
